from typing import Union, Optional

from decouple import config

from pysettings_yaml.providers.interfaces import SettingsProvider, NoValue, OriginModel


class DecoupleSettingsProvider(SettingsProvider):
    name = "env"

    def get(
        self, setting_name: str, origin: OriginModel
    ) -> Union[Optional[str], NoValue]:
        return config(setting_name, default=NoValue())



from typing import Union, Optional

from pysettings_yaml.providers.interfaces import SettingsProvider, NoValue, OriginModel


class DirectValueModel(OriginModel):
    value: str


class DirectValueSettingsProvider(SettingsProvider):
    name = "direct"
    schema = DirectValueModel

    def get(
        self, setting_name: str, origin_data: DirectValueModel
    ) -> Union[Optional[str], NoValue]:
        return origin_data.value



from typing import Dict

from deepmerge import Merger

default_merger = Merger(
    # pass in a list of tuple, with the
    # strategies you are looking to apply
    # to each type.
    [(dict, ["merge"])],
    # next, choose the fallback strategies,
    # applied to all other types:
    ["override"],
    # finally, choose the strategies in
    # the case where the types conflict:
    ["override"],
)


def merge(*dicts: Dict) -> Dict:
    if len(dicts) < 2:
        raise Exception("Need at least two dicts to merge")

    base: Dict = {}
    for dict in dicts:
        default_merger.merge(base, dict)

    return base



from abc import ABC, abstractmethod
from typing import Union, Optional, Dict

from pydantic import BaseModel


class NoValue:
    pass


class OriginModel(BaseModel):
    name: str


class SettingsProvider(ABC):
    @staticmethod
    @property
    @abstractmethod
    def name() -> str:
        pass

    schema = OriginModel

    def parse_option(
        self, setting_name: str, origin_data: Optional[Dict] = None
    ) -> Union[Optional[str], NoValue]:
        origin = self.schema(**origin_data) if origin_data else self.schema(name="N/A")

        return self.get(setting_name, origin)

    @abstractmethod
    def get(
        self, setting_name: str, origin_data: OriginModel
    ) -> Union[Optional[str], NoValue]:
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysettings_yaml', 'pysettings_yaml.providers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'deepmerge>=1.1.0,<2.0.0',
 'funcy>=1.17,<2.0',
 'pydantic>=1.10.2,<2.0.0',
 'python-decouple>=3.6,<4.0',
 'split-settings>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'pysettings-yaml',
    'version': '0.1.2',
    'description': 'Settings library for human beings',
    'long_description': '# Pysettings YAML\n\nThis library extends over the concepts of [Decouple](https://github.com/henriquebastos/python-decouple/) \nand [Split Settings](https://django-split-settings.readthedocs.io/en/latest/)\nto create a library that is simple to use and expresses your settings in a\nyaml file for higher expressivity.\n\n## Installation\n\n```\npip install pysettings-yaml\n```\n\n## Usage\n\nDefine a YAML file with the definition of your settings, for example, \na `settings.yaml` in the same path as your settings module:\n\n```yaml\nsettings:\n  SAMPLE_SETTING_BOOL:  # Each entry inside the settings section is the name of your setting\n    origins:\n      - name: env\n      - name: direct\n        value: true\n  SAMPLE_SETTING_STR:\n    origins:\n      - name: direct\n        value: banana\n      - name: env\n```\n\nThen you can use this file like this in a `settings.py`:\n\n```python\nfrom split_settings.tools import optional\nfrom pathlib import Path\n\nfrom pysettings_yaml import get_config\n\nBASE_DIR = Path(__file__).parent\n\nsetting_files = [\n    BASE_DIR / "settings.yaml",\n]\n\n\nconfig = get_config(setting_files)\n\nSAMPLE_SETTING_BOOL = config("SAMPLE_SETTING_BOOL", cast=bool)\nSAMPLE_SETTING_STR = config("SAMPLE_SETTING_STR")\n\nprint(SAMPLE_SETTING_BOOL)\nprint(SAMPLE_SETTING_STR)  # This will print banana\n```\n\nBy default the library supports two origin names: `env` (to receive the value of your setting using the _decouple_ library)\nor `direct` (to provide a constant value), but you can add your own providers easily\n(refer to the "Adding custom providers" section).\n\nOrigin are evaluated in strict order. In those examples, `SAMPLE_SETTING_BOOL`\nwill evaluate first the `env` origin and, if that value is `None`, it will\nprocess the next origin, which would be `direct`. In contrast, for `SAMPLE_SETTING_STR`,\nthe first origin is the `direct` origin. Since direct can never return `None`,\nthe env origin will never be evaluated.\n\n\n## Overriding settings\n\nYou can define several setting files to be processed in order.\nAll of them will be merged together, so any subsequent setting file\nwill override the settings with the same name defined in a previous file\nof the list. Very useful if you need to define different origins\nfor your settings depending on your deployment environment:\n\n```python\nfrom split_settings.tools import optional\nfrom pathlib import Path\nimport os\n\nfrom pysettings_yaml import get_config\nENVIRONMENT = os.environ.get("ENVIRONMENT", "dev")\n\nBASE_DIR = Path(__file__).parent\n\nsetting_files = [\n    BASE_DIR / "settings.yaml",\n    BASE_DIR / f"settings.{ENVIRONMENT}.yaml",\n    optional(BASE_DIR / "settings.nonexistent.yaml")\n]\n\n\nconfig = get_config(setting_files)\n\nSAMPLE_SETTING_BOOL = config("SAMPLE_SETTING_BOOL", cast=bool)\nSAMPLE_SETTING_STR = config("SAMPLE_SETTING_STR")\n```\n\nAny path wrapped with the `optional` function will not raise an\nexception should that file not exist in the path you specify.\n\n## Adding custom providers\n\nIf you need other providers (for example for getting your settings from a vault\nlike Amazon SSM) you can implement your own providers like this:\n\n```python\nfrom typing import Union, Optional\n\nfrom pysettings_yaml.providers.interfaces import SettingsProvider, OriginModel, NoValue\n\n\nclass AWSModel(OriginModel):\n    path: str\n    decrypt: bool\n\n\nclass SampleAWSSettingsProvider(SettingsProvider):\n    name = "aws"\n    schema = AWSModel\n\n    def get(\n        self, setting_name: str, origin_data: AWSModel\n    ) -> Union[Optional[str], NoValue]:\n        return f"path: {origin_data.path}, decrypt: {origin_data.decrypt}"\n\n```\n\nJust make sure to:\n* Provide the `name` and a pydantic `schema`\n* Implement the `get` method\n\nThen you can obtain your config providing your `SettingProvider` and use it in your yaml:\n\n```python\nconfig = get_config(setting_files, additional_providers=[SampleAWSSettingsProvider()])\n```\n\n```yaml\nsettings:\n  SAMPLE_SETTING:\n    origins:\n      - name: aws\n        path: some/path\n        decrypt: true\n      - name: env\n```\n',
    'author': 'David',
    'author_email': 'davigetto@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Rydra/pysettings-yaml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import os
from pathlib import Path
from typing import Optional, List, Dict, Union, Any, cast

from decouple import undefined, Undefined, UndefinedValueError, strtobool

import yaml
from pydantic import BaseModel
from split_settings.tools import _Optional

from pysettings_yaml.providers.decouple import DecoupleSettingsProvider
from pysettings_yaml.providers.direct import DirectValueSettingsProvider
from pysettings_yaml.providers.interfaces import SettingsProvider, NoValue
from pysettings_yaml.utils import merge

logger = logging.getLogger()


def _cast_boolean(value: Any) -> bool:
    """
    Helper to convert config values to boolean as ConfigParser do.
    """
    value = str(value)
    return bool(value) if value == "" else bool(strtobool(value))


def _cast_do_nothing(value: Any) -> Any:
    return value


def cast_value(value: Any, cast: Any = undefined) -> Any:
    if isinstance(cast, Undefined):
        cast = _cast_do_nothing
    elif cast is bool:
        cast = _cast_boolean

    return cast(value)


class SettingsRepository:
    def __init__(
        self, setting_providers: Dict[str, SettingsProvider], registry: "RegistryModel"
    ) -> None:
        self.setting_providers = setting_providers
        self.registry = registry

    def get_config_value(
        self,
        setting_name: str,
        default: Any = undefined,
        cast: Any = undefined,
    ) -> Optional[Any]:
        if setting_name in self.registry.settings:
            # If there is no registry, only the ConfigSettingsGetter can be used
            setting_metadata = self.registry.settings[setting_name]
            origins = setting_metadata.origins

            value: Union[NoValue, Any] = NoValue()
            for origin in origins:
                name = origin.get("name")
                if not name:
                    raise ValueError(
                        "Any definition of origin must contain at least the name property"
                    )
                getter = self.setting_providers.get(name)
                if not getter:
                    raise ValueError(f"Unknown origin {name}")

                value = getter.parse_option(
                    setting_name=setting_name, origin_data=origin
                )
                if not isinstance(value, NoValue):
                    break
        else:
            # If there is no registry for the given settings,
            # only the ConfigSettingsGetter can be used
            value = DecoupleSettingsProvider().parse_option(setting_name=setting_name)

        if isinstance(value, NoValue):
            if isinstance(default, Undefined):
                raise UndefinedValueError(
                    "{} not found. Declare it as envvar or define a default value.".format(
                        setting_name
                    )
                )

            value = default

        return cast_value(value, cast)


class RegistryModel(BaseModel):
    class SettingsModel(BaseModel):
        origins: List[Dict]

    settings: Dict[str, SettingsModel]


def get_config(
    setting_paths: Optional[List[Union[str, Path, _Optional]]] = None,
    registry: Optional[RegistryModel] = None,
    additional_providers: Optional[List[SettingsProvider]] = None,
) -> Any:
    """

    :param setting_paths: a list of `Path` or strings that point to the
    path of the yaml files. This is ignored if registry is supplied
    :param registry: a registry that contains all the data
    that defines where and how to obtain a setting
    :param additional_providers: a list of additional `SettingsProvider` you
    would like to include
    :return:
    """

    setting_paths = setting_paths or []
    if not setting_paths and not registry:
        raise ValueError(
            "Either the settings_paths or the registry variables must be set"
        )

    additional_providers = additional_providers or []
    setting_providers: Dict[str, SettingsProvider] = dict(
        (cast(str, p.name), p)
        for p in [
            *additional_providers,
            DecoupleSettingsProvider(),
            DirectValueSettingsProvider(),
        ]
    )

    return SettingsRepository(
        setting_providers, registry or load_registry(setting_paths)
    ).get_config_value


def load_registry(setting_paths: List[Union[str, _Optional]]) -> RegistryModel:
    registry: Dict = {}
    for path in setting_paths:
        if os.path.exists(path):
            with open(path) as fs:
                registry = merge(registry, yaml.safe_load(fs))
        elif not isinstance(path, _Optional):
            raise IOError("No such file: {0}".format(path))

    return RegistryModel(**registry)
